# This script is covered by the terms given in upcxx/LICENSE.txt
# The point of this script is to place compilers in your $PATH.
# If you're using Docker or self-install, you can probably skip this script.
# Note tcsh graciously yields 'BASH_VERSION: Undefined variable' before giving up here.
if test -z "$BASH_VERSION$ZSH_VERSION"; then
  echo 'WARNING: Your shell is not bash or zsh.'
  echo '         This script has not been tested with other bourne shells'
fi
echo Adding compilers to your PATH...
today=`date '+%D'`
#TESTING: today='07/26/23'
function try_set() {
  local var="$1"
  local val="$2"
  if [[ -z ${!var} ]]; then
    eval export $var="$val"
  elif [[ ${!var} != $val ]]; then
    echo "WARNING: Refusing to set $var='$val'."
    echo "         If the current setting $var='${!var}' is not correct,"
    echo "         then you should manually set it as appropriate."
  fi
}
if [[ $NERSC_HOST == perlmutter ]]; then
  echo "Host identified as NERSC Perlmutter"
  defacct=$(sacctmgr -p -n show user name=$USER | cut -d\| -f2)
  if (($?)); then
    unset defacct
  fi

  export TMPDIR='/dev/shm'
  export SLURM_THREADS_PER_CORE=1
  ml -q unload upcxx upcxx-cuda
  ml -q PrgEnv-gnu cpu gcc/11.2.0
  ml contrib upcxx/2023.3.0
  unset UPCXX_INSTALL
  export PMIRUN_CMD='srun -K0 -W60 %V -n %N -- %C'
  ml opencoarrays
  ml chapel cray-pmi
  export CHPL_RT_MAX_HEAP_SIZE="16GB"
  export CHPL_LAUNCHER_MEM=unset

  if [[ $today =~ 07/2[3-5]/23 ]]; then
    echo "WARNING: TESTING pre-reservation logic"
    export SLURM_CONSTRAINT=cpu
    unset SLURM_RESERVATION
  fi

  # Reservation for CPU-only nodes July 26 and 27, 2023
  if [[ $today =~  07/2[6-7]/23 ]]; then
    echo "Event identified as CUF23"
    try_set SLURM_CONSTRAINT cpu
    try_set SLURM_ACCOUNT ntrain7
  fi
  if [[ $today == 07/26/23 ]]; then
    try_set SLURM_RESERVATION cuf23_day1
  elif [[ $today == 07/27/23 ]]; then
    [[ $SLURM_RESERVATION == cuf23_day1 ]] && unset SLURM_RESERVATION
    try_set SLURM_RESERVATION cuf23_day2
  fi
  if [[ -n $SLURM_RESERVATION ]]; then
    export SALLOC_RESERVATION=$SLURM_RESERVATION
  fi

  # Always need a default time limit
  try_set SLURM_TIMELIMIT '00:10:00'

  # Always need a default account
  if [[ -z $SLURM_ACCOUNT && -n $defacct ]]; then
    echo "WARNING: setting SLURM_ACCOUNT='$defacct'"
    echo '         If this is not the correct account to charge,'
    echo '         then you should manually set it as appropriate.'
    export SLURM_ACCOUNT="$defacct"
  fi
  if [[ -z $SLURM_ACCOUNT ]]; then
    echo 'WARNING: The environment variable SLURM_ACCOUNT is unset or empty.'
    echo '         To launch applications as described in the tutorial, one must'
    echo '         set SLURM_ACCOUNT to an appropriate NERSC project id.'
    echo '         example: `export SLURM_ACCOUNT=m9999`'
    echo '         The `iris` command can be use to determine your project membership.'
  else
    export SALLOC_ACCOUNT=$SLURM_ACCOUNT
  fi
elif [[ $LMOD_SYSTEM_NAME == frontier ]]; then
  echo "Host identified as OLCF Frontier"
  defacct=$(sacctmgr -p -n show user name=$USER | cut -d\| -f2)
  if (($?)); then
    unset defacct
  fi

  export TMPDIR='/dev/shm'
  ml -q PrgEnv-gnu gcc/11.2.0 cray-pmi
  ml -q ums ums014 upcxx/2023.3.0
  unset UPCXX_INSTALL
  ml opencoarrays
  ml chapel-cuf23 cray-pmi
  export CHPL_RT_MAX_HEAP_SIZE="16GB"
  export CHPL_LAUNCHER_MEM=unset
  export CHPL_LAUNCHER_CORES_PER_LOCALE=56

  if [[ $today =~ 07/2[3-5]/23 ]]; then
    echo "WARNING: TESTING pre-reservation logic"
    unset SLURM_RESERVATION
  fi

  # Reservation for July 26 and 27, 2023
  if [[ $today =~ 07/2[67]/23 ]]; then
    echo "Event identified as CUF23"
    try_set SLURM_RESERVATION pgas_training
  fi
  if [[ -n $SLURM_RESERVATION ]]; then
    export SALLOC_RESERVATION=$SLURM_RESERVATION
  fi

  # Always need a default time limit
  try_set SLURM_TIMELIMIT '00:10:00'

  # Always need a default account
  if [[ -z $SLURM_ACCOUNT && -n $defacct ]]; then
    echo "WARNING: setting SLURM_ACCOUNT='$defacct'"
    echo '         If this is not the correct account to charge,'
    echo '         then you should manually set it as appropriate.'
    export SLURM_ACCOUNT="$defacct"
  fi
  if [[ -z $SLURM_ACCOUNT ]]; then
    echo 'WARNING: The environment variable SLURM_ACCOUNT is unset or empty.'
    echo '         To launch applications as described in the tutorial, one must'
    echo '         set SLURM_ACCOUNT to an appropriate OLCF project id.'
    echo '         example: `export SLURM_ACCOUNT=ABC123`'
    echo '         The `showusage` command can be use to determine your project membership.'
  else
    export SALLOC_ACCOUNT=$SLURM_ACCOUNT
  fi
elif [[ $(id -un) == 'tutorial' ]] && ( type -p aws &>/dev/null ); then
  echo "Host identified as AWS"
else
  echo 'This SOURCE_ME.sh script did not recognize your host.'
fi
for x in ftdetect syntax; do
  dir=".vim/$x"
  file="$dir/chpl.vim"
  if [[ ! -e ~/$file ]]; then
    echo "Creating ~/$file for Chapel syntax highlighting"
    mkdir -p ~/$dir && cp chapel/$file ~/$dir/
  fi
done
found=''
missing=''
for tool in upcxx chpl caf ; do
  if  type $tool &> /dev/null; then
    found+=" $tool"
  else
    missing+=" $tool"
  fi
done
if [[ -z $missing ]]; then
  echo 'Congratulations, the required tools are in your $PATH.'
else
  echo 'WARNING: The following tool(s) were found in your $PATH:'
  echo "           $found"
  echo '         but the following were NOT found:'
  echo "           $missing"
  echo '         If you need the missing tool(s), then you will need '
  echo '         to manually add their locations to your $PATH.'
fi

echo
echo "For UPC++ examples: cd upcxx  ;  more README.md"
echo "For Chapel examples: cd chapel ; more README.md"
echo "For Fortran examples: cd fortran ; more README.md"
echo
